<?php

define('RBS_PROD_URL' , 'https://securepayments.berekebank.kz/payment/rest/');
define('RBS_TEST_URL', 'https://3dsec.berekebank.kz/payment/rest/');

define('RBS_ENABLE_LOGGING' , true);
define('RBS_MODULE_VERSION', '2.1.3');
define('RBS_DEFAULT_MEASURE_NAME' , 'шт');
define('RBS_DEFAULT_MEASURE_CODE' , 0);

define('RBS_ENABLE_CALLBACK', true);


?><?php
use \UmiCms\Service;

class RBSPaymentPayment extends payment
{

    /**
     * @var array $validTaxIdList Yandex => RBSPayment (will not create unnecessary code)
     */
    private $validTaxIdList = array(
        2 => 0, 3 => 2, 4 => 3
    );

    private $_ffd_paymentMethodType;
    private $_ffd_paymentObjectType;
    private $_ffd_version;

    /**
     * {@inheritdoc}
     */
    public static function getOrderId()
    {
        $p = explode("_", getRequest('orderNumber'));
        return (int)$p[0];
    }

    /**
     * {@inheritdoc}
     */
    public function validate()
    {
        return true;
    }

    /**
     * {@inheritdoc}
     * Set order state to initialized
     */
    public function process($template = null)
    {
        $order = $this->order;
        $object = $this->object;

        /**
         * @var emarket $emarket
         */
        $cmsController = cmsController::getInstance();
        $emarket = $cmsController->getModule('emarket');
        /**
         * @var iUmiObject $currency
         */
        $currency = $emarket->getDefaultCurrency();
        $currency = ($currency instanceof iUmiObject) ? $currency->getValue('codename') : 'USD';

        if ($currency == 'USD') {
            $config['currency'] = 840;
        }
//        if (!strlen($object->rbspayment_user_name) || !strlen($object->rbspayment_merchant_pasword)) {
//            throw new publicException(getLabel('error-payment-wrong-settings'));
//        }
        $order_number = $this->order->getId();
        $amount = round($this->order->getActualPrice() * 100);

        $config['user_name'] = $object->rbspayment_user_name;
        $config['password'] = $object->rbspayment_password;
        $config['two_stage'] = $object->rbspayment_two_stage;
        $config['test_mode'] = $object->rbspayment_test_mode;
        $config['send_order'] = $object->rbspayment_send_order;

        $config['tax_system'] = $this->_getValueByObjectIdName($object->rbspayment_tax_system);
        $config['ffd_version'] = $this->_ffd_version = $this->_getValueByObjectIdName($object->rbspayment_FFDVersion);
        $config['ffd_paymentMethodType'] = $this->_ffd_paymentMethodType = $this->_getValueByObjectIdName($object->rbspayment_payment_methodtype);
        $config['ffd_paymentObjectType'] = $this->_ffd_paymentObjectType = $this->_getValueByObjectIdName($object->rbspayment_payment_objecttype);

        $config['logging'] = RBS_ENABLE_LOGGING;

        $registrySettings = Service::RegistrySettings();
        $config['cms_name'] = "UMI";
        $config['cms_version'] = "version: " . $registrySettings->getVersion() . " revision: " . $registrySettings->getRevision();
        $config['module_version'] = RBS_MODULE_VERSION;

        // collect orderBundle
        if ($config['send_order']) {
            $config['order_bundle'] = $this->getOrderBundle($order);

            // DISCOUNT CALCULATE
            $discountHelper = new DiscountHelper();
            $discount = $discountHelper->discoverDiscount($amount, $config['order_bundle']['cartItems']['items']);

            if($discount > 0) {
                $discountHelper->setOrderDiscount($discount);
                $recalculatedPositions = $discountHelper->normalizeItems($config['order_bundle']['cartItems']['items']);
                $amount = $discountHelper->getResultAmount();
                $config['order_bundle']['cartItems']['items'] = $recalculatedPositions;
            }
        }

        $rbsApi = new RBSPayment($config);

        $return_url = 'http://' . cmsController::getInstance()->getCurrentDomain()->getHost() . '/emarket/gateway/' . $this->order->getId();

        $response = $rbsApi->register_order($order_number, $amount, $return_url);

        $errorCode = empty($response['errorCode']) ? 0 :$response['errorCode'] ;

        if ($errorCode == 0) {
            $params = array();
            $params['formAction'] = $response['formUrl'];
            $params['mdOrder'] = $response['orderId'];

            $this->order->setPaymentStatus('initialized');

            list($templateString) = emarket::loadTemplates(
                'emarket/payment/rbs_payment/' . $template,
                'form_block',
                'mode_type_item'
            );
            // immediate redirect to payment form (without template)
//            header("Location: " . $params['formAction']);
            return emarket::parseTemplate($templateString, $params);

        } else {
            throw new privateException("ERROR #{$response['errorCode']}: {$response['errorMessage']}");
        }
    }

    public function _getValueByObjectIdName($objectId)
    {
        $objectsCollection = umiObjectsCollection::getInstance();
        $gettedObjC = $objectsCollection->getObject($objectId);
        $str = $gettedObjC->getName();
        return (int)filter_var($str, FILTER_SANITIZE_NUMBER_INT);
    }

    protected function getOrderBundle(order $order)
    {

        $customerId = $order->getCustomerId();
        $customerSource = selector::get('object')
            ->id($customerId);

        if (!$customerSource instanceof iUmiObject) {
            throw new publicException(getLabel('error-payment-wrong-customer-id'));
        }
        $customer = new customer($customerSource);

        $date = new DateTime();

        $order_bundle = array(
            'orderCreationDate' => $date->getTimestamp(),
            'customerDetails' => array(
                'email' => $this->getCustomerEmail($customer)
            ),
            'cartItems' => array('items' => $this->getOrderItemInfoList($order))
        );


        if (!is_array($order_bundle)) {
            throw new publicException(getLabel('error-payment-wrong-receipt-info'));
        }

        return $order_bundle;
    }

    protected function getCustomerEmail(customer $customer)
    {

        $email = $customer->getEmail();

        if (!is_string($email) || empty($email) || !umiMail::checkEmail($email)) {
            throw new publicException(getLabel('error-payment-wrong-customer-email'));
        }
        return $email;
    }

    protected function getOrderItemInfoList(order $order)
    {
        $orderItemList = $order->getItems();

        if (!is_array($orderItemList) || empty($orderItemList)) {
            throw new publicException(getLabel('error-payment-empty-order'));
        }

        $orderItemDataList = [];

        $itemsCnt = 1;
        foreach ($orderItemList as $orderItem) {
            $orderItemDataList = $this->getOrderItemInfo($orderItem, $itemsCnt);

            // FFD 1.05 added
//            if ($this->_ffd_version == 'v1_05') {

                $attributes = array();
                $attributes[] = array(
                    "name" => "paymentMethod",
                    "value" => $this->_ffd_paymentMethodType
                );
                $attributes[] = array(
                    "name" => "paymentObject",
                    "value" => $this->_ffd_paymentObjectType
                );

                $orderItemDataList['itemAttributes']['attributes'] = $attributes;
//            }

            $items[] = $orderItemDataList;

            $itemsCnt++;
        }

        try {
            $deliveryInfo = $this->getDeliveryInfo($order, $itemsCnt);
            // FFD 1.05 added
//            if ($this->_ffd_version == 'v1_05') {

                $attributes = array();
                $attributes[] = array(
                    "name" => "paymentMethod",
                    "value" => $this->_ffd_paymentMethodType
                );
                $attributes[] = array(
                    "name" => "paymentObject",
                    "value" => 4
                );

                $deliveryInfo['itemAttributes']['attributes'] = $attributes;
//            }
            $items[] = $deliveryInfo;
        } catch (expectObjectException $e) {
            // noop
        }

        return $items;
    }

    public function getDefaultMeasurement()
    {
        if ($this->_ffd_version == "v1_05") {
            return RBS_DEFAULT_MEASURE_NAME;
        }
        return RBS_DEFAULT_MEASURE_CODE;
    }

    /**
     * @param orderItem $orderItem
     * @return array
     */
    protected function getOrderItemInfo(orderItem $orderItem, $itemsCnt)
    {

        $item_price = sprintf('%.2f', $orderItem->getActualPrice()) * 100;
        $item_quantity = sprintf('%.3f', $orderItem->getAmount());

        return [
            'positionId' => $itemsCnt,
            'name' => $orderItem->getName(),
            'itemAmount' => $item_price * $item_quantity,
            'itemPrice' => $item_price,
            'quantity' => array(
                'value' => sprintf('%.2f', $item_quantity),
                'measure' => $this->getDefaultMeasurement()
            ),
            'tax' => array(
                'taxType' => $this->getEntityTaxRateId($orderItem->getObject()),
            ),
            'itemCode' => $orderItem->getId(),

        ];
    }

    /**
     * @param iUmiObject $object
     * @return int
     * @throws publicException
     */
    protected function getEntityTaxRateId(iUmiObject $object)
    {
        $taxRate = selector::get('object')
            ->id($object->getValue('tax_rate_id'));

        if (!$taxRate instanceof iUmiObject) {
            return 0;
            //throw new publicException(getLabel('error-payment-order-item-empty-tax'));
        }

        $taxRateId = (int)$taxRate->getValue('yandex_id');

        if (!in_array($taxRateId, $this->validTaxIdList)) {
            return 0;
            // throw new publicException(getLabel('error-payment-order-item-empty-tax'));
        }

        return $taxRateId;
    }

    /**
     * @param order $order
     * @return array
     * @throws expectObjectException
     */
    protected function getDeliveryInfo(order $order, $itemsCnt)
    {
        $delivery = selector::get('object')->id(
            $order->getValue('delivery_id')
        );

        if (!$delivery instanceof iUmiObject) {
            throw new expectObjectException(getLabel('error-unexpected-exception'));
        }

        return [
            'positionId' => $itemsCnt,
            'itemPrice' => sprintf('%.2f', $order->getValue('delivery_price')) * 100,
            'itemAmount' => sprintf('%.2f', $order->getValue('delivery_price')) * 100,
            'quantity' => array(
                'value' => sprintf('%.2f', 1),
                'measure' => $this->getDefaultMeasurement()
            ),
            'tax' =>
                array(
                    'taxType' => $this->getEntityTaxRateId($delivery),
                ),
            'name' => $delivery->getName(),
            'itemCode' => $delivery->getId()
        ];
    }

    public function poll()
    {

        $buffer = outputBuffer::current();
        $buffer->clear();
        $buffer->contentType('text/plain');

        $config['user_name'] = $this->object->rbspayment_user_name;
        $config['password'] = $this->object->rbspayment_password;
        $config['test_mode'] = $this->object->rbspayment_test_mode;
        $config['logging'] = RBS_ENABLE_LOGGING;

        $rbsApi = new RBSPayment($config);

        $host = 'http://' . cmsController::getInstance()->getCurrentDomain()->getHost();

        if (!empty(getRequest('orderId'))) {
            $action = "result";
            $gatewayId = getRequest('orderId');
        }
        elseif (!empty(getRequest('mdOrder'))) {
            $action = "callback";
            $gatewayId = getRequest('mdOrder');
        }

        $response = $rbsApi->get_order_status_by_orderId($gatewayId);


        switch ($action) {

            case "result":
                //by returnUrl
                if ($response['ErrorCode'] == 0 && (($response['orderStatus'] == 1) || ($response['orderStatus'] == 2))) {
                    if (RBS_ENABLE_CALLBACK === false) {
                        $this->order->setPaymentStatus('accepted');
                    }
                    $host .= '/emarket/purchase/result/successful/';
                } else {
                    if (RBS_ENABLE_CALLBACK === false) {
                        $this->order->setPaymentStatus('failed');
                    }
                    $host .= '/emarket/purchase/result/fail/';
                }
                break;

            case "callback":
                //by callback
                if ($response['ErrorCode'] == 0 && (($response['orderStatus'] == 1) || ($response['orderStatus'] == 2))) {
                    $this->order->setPaymentStatus('accepted');
                    $host .= '/emarket/purchase/result/successful/';
                } else {
                    $this->order->setPaymentStatus('failed');
                    $host .= '/emarket/purchase/result/fail/';
                }
                break;
        }

        $cmsController = cmsController::getInstance();
        $module = $cmsController->getModule("emarket");
        if ($module) {
            $this->order->commit();
            $module->redirect($host);
        }
    }

}

/**
 * @author RBS
 * @version 2.1
 */
class RBS
{
    /**
     * @var string
     */
    public $cms_name;
    /**
     * @var string
     */
    public $cms_version;
    /**
     * @var string
     */
    public $module_version = '';

    /**
     * @var boolean
     */
    public $send_order;
    public $tax_system;
    public $order_bundle = [];
    public $language = 'en';
    public $logging;

    public $user_name;
    public $password;
    public $two_stage;
    public $test_mode;

    public function __construct($params = array())
    {
        foreach ($params as $key => $value) {
            $this->{$key} = $value;
        }
    }

    function register_order($order_number, $amount, $return_url)
    {
        $data = array(
            'orderNumber' => $order_number . "_" . time(),
            'amount' => $amount,
            'returnUrl' => $return_url,
            'language' => $this->language,
            'jsonParams' => json_encode(
                array(
                    'CMS:' => $this->cms_name . " " . $this->cms_version,
                    'Module-Version: ' => $this->module_version
                )
            ),
        );

        if ($this->two_stage) {
            $method = 'registerPreAuth.do';
        } else {
            $method = 'register.do';
        }

        if ($this->send_order && !empty($this->order_bundle)) {
            $data['orderBundle'] = json_encode($this->order_bundle);
            $data['taxSystem'] = $this->tax_system;
        }


        //CBLO
        if (RBS_ENABLE_CALLBACK === true) {

            $action_adr = $this->_getActionAddress();
            if ($this->test_mode == true) {
                $gate_url = str_replace("payment/rest", "mportal-uat/mvc/public/merchant/update", $action_adr);
                // "mportal-uat" error for some ONE bank
                //$gate_url = str_replace("rest", "mportal/mvc/public/merchant/update", $action_adr);

            } else {
                $gate_url = str_replace("payment/rest", "mportal/mvc/public/merchant/update", $action_adr);
            }
            $gate_url .= substr($this->user_name, 0, -4); // we guess username = login w/o "-api"
//            $callback_addresses_string = $data['returnUrl'] . '?a=cb';
            $callback_addresses_string = Service::DomainDetector()->detectUrl() . '/emarket/gateway/';
            $this->_updateGatewayCallback($this->user_name, $this->password, $gate_url, $callback_addresses_string);
        }

        $response = $this->gateway($method, $data);
        return $response;
    }

    private function _getActionAddress() {
        $action_adr = RBS_TEST_URL;
        if ($this->test_mode != TRUE) {
            $action_adr = RBS_PROD_URL;
            // for migrate to new payment gateway
            if (defined('RBS_PROD_URL_ALT') && defined('RBS_PROD_URL_ALT_PREFIX')) {
                if (substr($this->user_name, 0, strlen(RBS_PROD_URL_ALT_PREFIX)) == RBS_PROD_URL_ALT_PREFIX) {
                    $action_adr = RBS_PROD_URL_ALT;
                } else {
                    //disable callbacks for old gateway version
                }
            }
        }
        return $action_adr;
    }

    private function _updateGatewayCallback($login, $password, $action_address, $callback_addresses_string)
    {
        $headers = array(
            'Content-Type:application/json',
            'Authorization: Basic ' . base64_encode($login . ":" . $password)
        );
        $data['callbacks_enabled'] = true;
        $data['callback_type'] = "STATIC";
        $data['callback_addresses'] = $callback_addresses_string;
        $data['callback_operations'] = "deposited,approved,declinedByTimeout";
        $response = $this->_sendGatewayData(json_encode($data), $action_address, $headers);
        return $response;
    }

    private function _sendGatewayData($data, $action_address, $headers = array())
    {
        $ch = curl_init();
        curl_setopt_array($ch, array(
            CURLOPT_HTTPHEADER => $headers,
            CURLOPT_VERBOSE => true,
            CURLOPT_SSL_VERIFYHOST => false,
            CURLOPT_SSL_VERIFYPEER => false,
            CURLOPT_URL => $action_address,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_POST => true,
            CURLOPT_POSTFIELDS => $data,
            //                CURLOPT_ENCODING, "gzip",
        ));
        $response = curl_exec($ch);
        curl_close($ch);

        return $response;
    }

    private function gateway($method, $data)
    {
        $data['userName'] = $this->user_name;
        $data['password'] = $this->password;
        $data['language'] = $this->language;

        $action_adr = $this->_getActionAddress();
        $response = $this->_sendGatewayData(http_build_query($data, '', '&'), $action_adr . $method, array('CMS: ' . $this->cms_name . ' ' . $this->cms_version, 'Module-Version: ' . $this->module_version));
        $response = json_decode($response, true);

        if ($this->logging) {
            $this->rbs_logger('REQUEST: ' . $action_adr . $method . ': ' . json_encode($data) . ' RESPONSE: ' . json_encode($response));
        }
        return $response;
    }

    public function rbs_logger($var, $info = false)
    {
        $information = "";
        if ($var) {
            if ($info) {
                $information = "\n\n";
                $information .= str_repeat("-=", 64);
                $information .= "\nDate: " . date('Y-m-d H:i:s');
            }

            $result = $var;
            if (is_array($var) || is_object($var)) {
                $result = "\n" . print_r($var, true);
            }
            $result .= "\n\n";
            $path = dirname(__FILE__) . '/umi_rbspayment_' . date('Y-m') . '.log';

            error_log($information . $result, 3, $path);
            return true;
        }
        return false;
    }

    public function get_order_status_by_orderId($orderId)
    {
        $data = array('orderId' => $orderId);
        $response = $this->gateway('getOrderStatusExtended.do', $data);
        return $response;
    }

}
class RBSDiscount {

    private $cashBasket = array();
    private $equalPosition = -1;
    private $faultPrice;
    private $arrOrder = array(
        'order_id' =>  null,
        'amount' =>  null,
        'discount' =>  null,
        'basket' => array()
    );

    private $arrRecountSumm = array(
        'priceAmount' => 0,
        'priceDiscount' => 0,
        'discount' => 0,
    );

    private function priceTransform() {
        $this->arrOrder['discount'] = $this->arrOrder['discount'] * 100;
        foreach ($this->arrOrder['basket'] as $key => $item) {
            $item['discount'] = $item['discount'] * 100;
            $item['priceAmount'] = $item['priceAmount'] * 100;
            $item['priceBase'] = $item['priceBase'] * 100;
            $item['priceDiscount'] = $item['priceDiscount'] * 100;
            $this->arrOrder['basket'][$key] = $item;
        }
    }

    private function transformNumber($value, $count = 2) {
        $count = $count ? $count : 2;
        return round($value,$count);
    }

    private function transformBasePrice($item) {
        $result = ($this->arrOrder['discount'] * $item['percentDiscount'] / 100) / $item['count'];
        $result = $this->transformNumber( $item['priceBase'] - $result);
        if($result < 0) {  $result = 0; }
        return $result;
    }


    private function onPositionDiscountPercent($item) {
        $price = $item['priceAmount'];
        return $this->transformNumber((100 / $this->arrOrder['amount'] * $price), 4);
    }

    private function onPositionDiscount($item) {
        $result = $this->transformNumber( ($item['priceBase'] - $item['priceBaseDiscount']) * $item['count']);
        return $result;
    }

    private function onPositionDiscountPrice($item) {
        $result = $this->transformNumber($item['priceBaseDiscount'] * $item['count']);
        if( $result < 0 ) { $result = 0; }
        return $result;
    }


    private function findMaxCount() {
        $maxValue = max(array_column( $this->arrOrder['basket'], 'count' ));
        $result = array_search( $maxValue, array_column( $this->arrOrder['basket'] ,'count') );
        return $result;
    }

    private function splitPosition($mode, $indexSeparate) {
        if($mode === 'none') {
            return false;
        }
        $newIndex = count($this->arrOrder['basket']);
        $separatePosition = $this->arrOrder['basket'][$indexSeparate];

        if( $mode === 'onePosition' ) {
            $this->arrOrder['basket'][$indexSeparate]['count'] = $this->transformNumber( $separatePosition['count'] - 1,3 );
            array_push( $this->arrOrder['basket'], $separatePosition );
            $this->equalPosition = $newIndex;
            $this->arrOrder['basket'][$newIndex] = $separatePosition;
            $this->arrOrder['basket'][$newIndex]['count'] = 1;
        } else if( $mode === 'fractionPosition' ) {
            $needPositionPrice = $this->transformNumber( $separatePosition['priceDiscount'] + $this->faultPrice );
            $originalCount = $separatePosition['count'];
            $success = false;
            $vishPrice = $this->transformNumber( $needPositionPrice / $originalCount );

            $separateObject = array();
            $separateObject['needSumm'] = $needPositionPrice;
            $separateObject['oldPositionCount'] =  $this->transformNumber($originalCount / 2, 3 );
            $separateObject['newPositionCount'] = $this->transformNumber($originalCount - $separateObject['oldPositionCount'],3);


            for ($toIndex = 0; $toIndex < 5; $toIndex++) {
                $separateObject['oldPositionCount'] = $this->transformNumber( $separateObject['oldPositionCount'] - 0.001, 3) ;
                $separateObject['newPositionCount'] = $this->transformNumber( $separateObject['newPositionCount'] + 0.001, 3 );
                $separateObject['oldPositionPrice'] = $this->transformNumber( $separateObject['oldPositionCount'] * $vishPrice );
                $separateObject['newPositionPrice'] = $this->transformNumber( $separateObject['newPositionCount'] * $vishPrice );
                $separateObject['oldPositionVishPrice'] = $this->transformNumber( $separateObject['oldPositionPrice'] / $separateObject['oldPositionCount'] );
                $separateObject['newPositionVishPrice'] = $this->transformNumber( $separateObject['newPositionPrice'] / $separateObject['newPositionCount'] );
                $separateObject['recountSumm'] = $this->transformNumber( $separateObject['oldPositionPrice'] + $separateObject['newPositionPrice']);
                if($separateObject['recountSumm'] == $separateObject['needSumm']) {
                    $success = true;
                    break;
                }
            }
            if( !$success ) {
                $res = $this->transformNumber(($separateObject['newPositionPrice'] - ($separateObject['recountSumm']-$separateObject['needSumm'])) / $separateObject['newPositionCount']);
                $separateObject['newPositionPrice'] = $this->transformNumber( $res * $separateObject['newPositionCount'] );
                $separateObject['recountSumm'] = $this->transformNumber( $separateObject['oldPositionPrice']+$separateObject['newPositionPrice'] );
                if( $separateObject['recountSumm'] === $separateObject['needSumm'] ) { $success = true;}
            }
            if( $success ) {
                $separatePosition['count'] = $separateObject['oldPositionCount'];
                $separatePosition['priceBaseDiscount'] = $separateObject['oldPositionVishPrice'];
                $separatePosition['priceAmount'] = $this->transformNumber( $separatePosition['count'] * $separatePosition['priceBase'] );
                $separatePosition['priceDiscount'] = $separateObject['oldPositionPrice'];
                $separatePosition['discount'] = $this->transformNumber( $separatePosition['priceAmount'] - $separatePosition['priceDiscount'] );
                $this->arrOrder['basket'][$indexSeparate] = $separatePosition;
                array_push($this->arrOrder['basket'], $separatePosition);

                $newPosition = $this->arrOrder['basket'][$newIndex];
                $newPosition['count'] = $separateObject['newPositionCount'];
                $newPosition['priceAmount'] = $this->transformNumber( $newPosition['count'] * $newPosition['priceBase'] );
                $newPosition['priceBaseDiscount'] = $separateObject['newPositionVishPrice'];
                $newPosition['priceDiscount'] = $separateObject['newPositionPrice'];
                $newPosition['discount'] = $this->transformNumber( $newPosition['priceAmount'] - $newPosition['priceDiscount'] );
                $this->arrOrder['basket'][$newIndex] = $newPosition;

            } else {

            }

        }
    }

    private function transformBasket() {
        $onePositionFind = false;
        $positionSeparateIndex = -1;
        $basket = $this->arrOrder['basket'];
        $transformMode = 'none';
        $this->equalPosition = -1;

        if( count($basket) === 0 ) { return false; }

        foreach ($basket as $key => $position) {
            if($position['count'] > 1) {
                $positionSeparateIndex = $key;
            }
            if ($position['count'] == 1) {
                $this->equalPosition = $key;
                $onePositionFind = true;
            }
        }

        if(!$onePositionFind && $positionSeparateIndex >= 0) {
            $transformMode = 'onePosition';
        } else if(!$onePositionFind) {
            $positionSeparateIndex = $this->findMaxCount();
            $transformMode = 'fractionPosition';
        }
        $this->splitPosition( $transformMode,$positionSeparateIndex );
        return $transformMode;
    }

    private function generateReceiptAmount() {
        $summ = 0;
        foreach ($this->arrOrder['basket'] as $key => $item) {
            $summ += $this->transformNumber($item['priceBase'] * $item['count']);
        }
        $this->arrOrder['amount'] = $this->transformNumber($summ);
        if($this->arrOrder['amount'] < $this->arrOrder['discount']) {
            $this->arrOrder['discount'] = $this->arrOrder['amount'];
        }

    }

    private function generatePositionsDiscount() {
        $difference = $this->arrOrder['amount'] - $this->arrOrder['discount'];
        if( $difference > 1 ) {
            foreach ($this->arrOrder['basket'] as $key => $item) {
                $item['priceAmount'] = $this->transformNumber( $item['priceBase'] * $item['count'] );
                $item['percentDiscount'] = $this->onPositionDiscountPercent( $item );
                $item['priceBaseDiscount'] = $this->transformBasePrice( $item );
                $item['discount'] = $this->onPositionDiscount( $item );
                $item['priceDiscount'] = $this->onPositionDiscountPrice( $item );

                $this->arrOrder['basket'][$key] = $item;
            }
        } else {
            foreach ($this->arrOrder['basket'] as $key => $item) {
                $item['priceAmount'] = $this->transformNumber( $item['priceBase'] * $item['count'] );
                $item['percentDiscount'] = 0;
                $item['priceBaseDiscount'] = 0;
                $item['discount'] = $item['priceAmount'];
                $item['priceDiscount'] = $item['priceAmount'] - $item['discount'];

                $this->arrOrder['basket'][$key] = $item;
            }
        }
    }

    private function equalizePositionDiscount() {

        if( $this->equalPosition < 0 ) { return false; }
        $equalPosition = $this->arrOrder['basket'][$this->equalPosition];
        $baseSumm = $this->arrOrder['amount'] - $this->arrOrder['discount'];
        $recountSumm = $this->arrRecountSumm['priceDiscount'];
//        $remain = 0;

        if( $baseSumm != $recountSumm ) {
            $remain = $this->transformNumber($baseSumm - $recountSumm);
            $this->arrOrder['basket'][$this->equalPosition]['priceBaseDiscount'] = $this->transformNumber($equalPosition['priceBaseDiscount'] + $remain);
            $this->arrOrder['basket'][$this->equalPosition]['priceDiscount'] = $equalPosition['priceBaseDiscount'];
            $this->arrOrder['basket'][$this->equalPosition]['discount'] = $this->transformNumber($equalPosition['discount'] - $remain);
            $this->arrOrder['basket'][$this->equalPosition]['priceDiscount'] = $this->arrOrder['basket'][$this->equalPosition]['priceBaseDiscount'];
        }
    }

    private function provideProductsToOrder() {
        $this->arrOrder['basket'] = $this->cashBasket;
    }

    private function finalCheck() {
        $baseAmount = $this->transformNumber( $this->arrOrder['amount'] - $this->arrOrder['discount'] );
        $finalAmount = $this->arrRecountSumm['priceDiscount'];

        $this->faultPrice =  $this->transformNumber($baseAmount - $finalAmount);
        if($this->faultPrice != 0) {
            $transformMode = $this->transformBasket();
            if( $transformMode !== 'fractionPosition' ) {
                $this->generatePositionsDiscount();
                $this->recountReceiptAmount();
                $this->equalizePositionDiscount();
            }
            $this->recountReceiptAmount();
        }
    }

    public function addProduct($product) {
        $product['priceAmount'] = $product['count'] * $product['priceBase'];
        $this->cashBasket[] = $product;
    }

    public function discoverDiscount($amount,$positions) {
        $positionsSumm =  $this->transformNumber(array_sum(array_column($positions, 'itemAmount')), 0);
        $discount = $this->transformNumber($positionsSumm - $amount,0);
        return $discount;
    }

    public function recountReceiptAmount() {
        $summ = array(
            'priceAmount' =>  0,
            'priceDiscount' =>  0,
            'discount' => 0
        );
        foreach ($this->arrOrder['basket'] as $key => $item) {
            $summ['priceAmount'] += $item['priceAmount'];
            $summ['priceDiscount'] += $item['priceBaseDiscount'] * $item['count'];
            $summ['discount'] += $item['discount'];
        }
        $this->arrRecountSumm['priceAmount'] = $this->transformNumber( $summ['priceAmount'] );
        $this->arrRecountSumm['priceDiscount'] = $this->transformNumber( $summ['priceDiscount'] );
        $this->arrRecountSumm['discount'] = $this->transformNumber( $summ['discount'] );

    }

    public function getOrder() {
        return $this->arrOrder;
    }


    public function getProducts() {
        return $this->cashBasket;
    }
    public function getBasketResult() {
//        $order;
        foreach ($this->arrOrder['basket'] as $key => $item) {
            $order[$key] = array(
                'positionId' => $key+1,
                'name' => $item['name'],
                'quantity' => array(
                    'value' => $item['count'],
                    'measure' => $item['arrGate']['quantity']['measure'],
                ),
                'itemAmount' => $this->transformNumber($item['priceDiscount'] * 100,0),
                'itemCode' => $item['id'],
                'itemPrice' => $this->transformNumber($item['priceBaseDiscount'] * 100,0),
                'tax' => array(
                    'taxType' => $item['arrGate']['tax']['taxType'],
                ),
            );

            if(isset($item['itemAttributes'])) {
                $order[$key]['itemAttributes'] = $item['itemAttributes'];
            }
        }
        return $order;

    }
    public function getRecountSumm() {
        return $this->arrRecountSumm;
    }

    public function setOrderDiscount($value) {
        $this->arrOrder['discount'] = $this->transformNumber( $value / 100 );
    }

    public function updateOrder() {
        $this->provideProductsToOrder();
        $this->generateReceiptAmount();
        $this->generatePositionsDiscount();
        $this->recountReceiptAmount();
        $this->finalCheck();
    }

    public function calculateDiscount( $amount, $discount) {
        $result = $this->transformNumber( $discount * $amount / 100,0);
        return $result;
    }

    public function normalizeItems($gatePositions) {

        foreach ($gatePositions as $key => $item) {

            $arrProduct = array(
                'id' => $item['positionId'],
                'name' => $item['name'],
                'priceBase' => $item['itemPrice'] / 100,
                'count' => $item['quantity']['value'],
                'arrGate' => $item
            );
            $this->addProduct($arrProduct);

        }

        $this->updateOrder();

        $arrResult = array();
        foreach ($this->arrOrder['basket'] as $key => $item) {
            $arrResult[$key] = $item['arrGate'];
            $arrResult[$key]['positionId'] = $key+1;
            $arrResult[$key]['itemPrice'] = $this->transformNumber($item['priceBaseDiscount'] * 100,0);
            $arrResult[$key]['itemAmount'] = $this->transformNumber($item['priceDiscount'] * 100,0);
            $arrResult[$key]['quantity']['value'] = $item['count'];

            if(isset($item['itemAttributes'])) {
                $order[$key]['itemAttributes'] = $item['itemAttributes'];
            }
        }
        return $arrResult;
    }

    public function getResultAmount() {
        return $this->transformNumber($this->arrRecountSumm['priceDiscount'] * 100,0);
    }

    public function test($message = '') {
        echo "<pre>";
        print_r($this->arrOrder);
        print_r($this->arrRecountSumm);
        echo "</pre>";
    }

}